@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2002-2019 ProcessOne, SARL. All Rights Reserved.
+ * Copyright (C) 2002-2020 ProcessOne, SARL. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -1216,6 +1216,44 @@ static ERL_NIF_TERM get_negotiated_cipher_nif(ErlNifEnv *env, int argc,
     return enif_make_binary(env, &bin);
 }
 
+static ERL_NIF_TERM tls_get_peer_finished_nif(ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[]) {
+    state_t *state = NULL;
+    if (!enif_get_resource(env, argv[0], tls_state_t, (void *) &state))
+        return enif_make_badarg(env);
+
+	/* OpenSSL does not offer an API to directly get the length of the
+	 * expected TLS Finished message, so just do a dummy call to grab this
+	 * information to allow caller to do an allocation with a correct size.
+	 */
+    ERL_NIF_TERM bin;
+    size_t len = SSL_get_peer_finished(state->ssl, NULL, 0);
+    if (len == 0)
+        return ERR_T(enif_make_atom(env, "undefined"));
+    unsigned char *buf = enif_make_new_binary(env, len, &bin);
+    if (!buf)
+        return ERR_T(enif_make_atom(env, "enomem"));
+
+    (void) SSL_get_peer_finished(state->ssl, buf, len);
+    return OK_T(bin);
+}
+
+static ERL_NIF_TERM tls_get_finished_nif(ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[]) {
+    state_t *state = NULL;
+    if (!enif_get_resource(env, argv[0], tls_state_t, (void *) &state))
+        return enif_make_badarg(env);
+
+    ERL_NIF_TERM bin;
+    size_t len = SSL_get_finished(state->ssl, NULL, 0);
+    if (len == 0)
+        return ERR_T(enif_make_atom(env, "undefined"));
+    unsigned char *buf = enif_make_new_binary(env, len, &bin);
+    if (!buf)
+        return ERR_T(enif_make_atom(env, "enomem"));
+
+    (void) SSL_get_finished(state->ssl, buf, len);
+    return OK_T(bin);
+}
+
 static ErlNifFunc nif_funcs[] =
         {
                 {"open_nif",                  8, open_nif},
@@ -1227,7 +1265,9 @@ static ErlNifFunc nif_funcs[] =
                 {"get_certfile_nif",          1, get_certfile_nif},
                 {"clear_cache_nif",           0, clear_cache_nif},
                 {"invalidate_nif",            1, invalidate_nif},
-                {"get_negotiated_cipher_nif", 1, get_negotiated_cipher_nif}
+                {"get_negotiated_cipher_nif", 1, get_negotiated_cipher_nif},
+                {"tls_get_peer_finished_nif", 1, tls_get_peer_finished_nif},
+                {"tls_get_finished_nif",      1, tls_get_finished_nif}
         };
 
 ERL_NIF_INIT(fast_tls, nif_funcs, load, NULL, NULL, unload)
