@@ -5,7 +5,7 @@
 %%% Created : 24 Jul 2004 by Alexey Shchepin <alexey@process-one.net>
 %%%
 %%%
-%%% Copyright (C) 2002-2019 ProcessOne, SARL. All Rights Reserved.
+%%% Copyright (C) 2002-2020 ProcessOne, SARL. All Rights Reserved.
 %%%
 %%% Licensed under the Apache License, Version 2.0 (the "License");
 %%% you may not use this file except in compliance with the License.
@@ -22,29 +22,23 @@
 %%%----------------------------------------------------------------------
 
 -module(fast_tls).
+-on_load(load_nif/0).
 
 -author('alexey@process-one.net').
 
--compile({no_auto_import, [{integer_to_binary, 1}]}).
-
--behaviour(gen_server).
-
 -export([open_nif/8, loop_nif/4, get_peer_certificate_nif/1,
          get_verify_result_nif/1, invalidate_nif/1,
          get_negotiated_cipher_nif/1]).
 
--export([start_link/0, tcp_to_tls/2,
+-export([tcp_to_tls/2,
          tls_to_tcp/1, send/2, recv/2, recv/3, recv_data/2,
          setopts/2, sockname/1, peername/1,
          controlling_process/2, close/1,
          get_peer_certificate/1, get_peer_certificate/2,
          get_verify_result/1, get_cert_verify_string/2,
          add_certfile/2, get_certfile/1, delete_certfile/1,
-         clear_cache/0, get_negotiated_cipher/1]).
-
-%% Internal exports, call-back functions.
--export([init/1, handle_call/3, handle_cast/2,
-         handle_info/2, code_change/3, terminate/2]).
+         clear_cache/0, get_negotiated_cipher/1,
+         get_tls_last_message/2]).
 
 -ifdef(TEST).
 -include_lib("eunit/include/eunit.hrl").
@@ -82,18 +76,6 @@
 
 -export_type([tls_socket/0]).
 
-start_link() ->
-    gen_server:start_link({local, ?MODULE}, ?MODULE, [],
-                          []).
-
-init([]) ->
-    case load_nif() of
-        ok ->
-            {ok, []};
-        {error, Why} ->
-            {stop, Why}
-    end.
-
 open_nif(_Flags, _CertFile, _Ciphers, _ProtocolOpts, _DHFile, _CAFile, _SNI, _ALPN) ->
     erlang:nif_error({nif_not_loaded, ?MODULE}).
 
@@ -124,36 +106,25 @@ clear_cache_nif() ->
 get_negotiated_cipher_nif(_Port) ->
     erlang:nif_error({nif_not_loaded, ?MODULE}).
 
+tls_get_peer_finished_nif(_Port) ->
+    erlang:nif_error({nif_not_loaded, ?MODULE}).
+
+tls_get_finished_nif(_Port) ->
+    erlang:nif_error({nif_not_loaded, ?MODULE}).
+
 %%% --------------------------------------------------------
 %%% The call-back functions.
 %%% --------------------------------------------------------
 
-handle_call(_, _, State) -> {noreply, State}.
-
-handle_cast(_, State) -> {noreply, State}.
-
-handle_info({'EXIT', Port, Reason}, Port) ->
-    {stop, {port_died, Reason}, Port};
-handle_info({'EXIT', _Pid, _Reason}, Port) ->
-    {noreply, Port};
-handle_info(_, State) -> {noreply, State}.
-
-code_change(_OldVsn, State, _Extra) -> {ok, State}.
-
-terminate(_Reason, _State) ->
-    ok.
-
--spec tcp_to_tls(inet:socket(),
-                 [{atom(), any()}]) -> {'error', 'no_certfile' | binary()} |
-{ok, tls_socket()}.
-
+-spec tcp_to_tls(inet:socket(), [atom() | {atom(), any()}]) ->
+    {ok, tls_socket()} | {'error', 'no_certfile' | binary()}.
 tcp_to_tls(TCPSocket, Options) ->
     Command = case lists:member(connect, Options) of
                   true -> ?SET_CERTIFICATE_FILE_CONNECT;
                   false -> ?SET_CERTIFICATE_FILE_ACCEPT
               end,
     CertFile = proplists:get_value(certfile, Options, ""),
-    if CertFile /= [] orelse Command == ?SET_CERTIFICATE_FILE_ACCEPT ->
+    if CertFile /= [] orelse Command == ?SET_CERTIFICATE_FILE_CONNECT ->
         Flags1 = case lists:member(verify_none, Options) of
                      true -> ?VERIFY_NONE;
                      false -> 0
@@ -211,32 +182,27 @@ tcp_to_tls(TCPSocket, Options) ->
     end.
 
 -spec tls_to_tcp(tls_socket()) -> inet:socket().
-
-tls_to_tcp(#tlssock{tcpsock = TCPSocket,
-                    tlsport = Port}) ->
+tls_to_tcp(#tlssock{tcpsock = TCPSocket, tlsport = Port}) ->
     invalidate_nif(Port),
     TCPSocket.
 
+-spec recv(tls_socket(), non_neg_integer()) ->
+    {error, inet:posix()} | {error, binary()} | {ok, binary()}.
 recv(Socket, Length) -> recv(Socket, Length, infinity).
 
--spec recv(tls_socket(), non_neg_integer(),
-           timeout()) -> {error, inet:posix()} |
-{error, binary()} |
-{ok, binary()}.
-
+-spec recv(tls_socket(), non_neg_integer(), timeout()) ->
+    {error, inet:posix()} | {error, binary()} | {ok, binary()}.
 recv(TLSSock, Length, Timeout) ->
     recv_and_loop(TLSSock, <<>>, <<>>, <<>>,
                   case Length of 0 -> -1; _ -> Length end, Timeout).
 
--spec recv_data(tls_socket(), binary()) -> {error, inet:posix() | binary()} |
-{ok, binary()}.
-
+-spec recv_data(tls_socket(), binary()) ->
+    {error, inet:posix() | binary()} | {ok, binary()}.
 recv_data(TLSSock, Packet) ->
     loop(TLSSock, <<>>, Packet, <<>>, -1).
 
 -spec loop(tls_socket(), binary(), binary(), binary(), integer()) ->
-    {error, inet:posix() | binary()} |
-    {ok, binary()}.
+    {error, inet:posix() | binary()} | {ok, binary()}.
 loop(#tlssock{tcpsock = TCPSocket,
               tlsport = Port} = Socket,
      ToSend, Received, DecBuf, Length) ->
@@ -257,9 +223,8 @@ loop(#tlssock{tcpsock = TCPSocket,
         {error, einval}
     end.
 
--spec recv_and_loop(tls_socket(), binary(), binary(), binary(), integer(), integer()) ->
-    {error, inet:posix() | binary()} |
-    {ok, binary()}.
+-spec recv_and_loop(tls_socket(), binary(), binary(), binary(), integer(), timeout()) ->
+    {error, inet:posix() | binary()} | {ok, binary()}.
 recv_and_loop(#tlssock{tcpsock = TCPSocket} = Socket,
               ToSend, Received, DecBuf, Length, Timeout) ->
     case loop(Socket, ToSend, Received, DecBuf, Length) of
@@ -287,8 +252,8 @@ recv_and_loop(#tlssock{tcpsock = TCPSocket} = Socket,
             end
     end.
 
--spec send(tls_socket(), binary()) -> ok | {error, inet:posix() | binary() | timeout}.
-
+-spec send(tls_socket(), binary()) ->
+    ok | {error, inet:posix() | binary() | timeout}.
 send(Socket, Packet) ->
     case loop(Socket, Packet, <<>>, <<>>, 0) of
         {ok, <<>>} ->
@@ -299,14 +264,13 @@ send(Socket, Packet) ->
             Other
     end.
 
--spec setopts(tls_socket(), list()) -> ok | {error, inet:posix()}.
-
+-spec setopts(tls_socket(), list()) ->
+    ok | {error, inet:posix()}.
 setopts(#tlssock{tcpsock = TCPSocket}, Opts) ->
     inet:setopts(TCPSocket, Opts).
 
--spec sockname(tls_socket()) -> {ok, {inet:ip_address(), inet:port_number()}} |
-{error, inet:posix()}.
-
+-spec sockname(tls_socket()) ->
+    {ok, {inet:ip_address(), inet:port_number()}} | {error, inet:posix()}.
 sockname(#tlssock{tcpsock = TCPSocket}) ->
     inet:sockname(TCPSocket).
 
@@ -351,8 +315,13 @@ get_negotiated_cipher(#tlssock{tlsport = Port}) ->
             error
     end.
 
--spec get_verify_result(tls_socket()) -> byte().
+-spec get_tls_last_message(peer | self, tls_socket()) -> {ok, binary()} | {error, term()}.
+get_tls_last_message(peer, #tlssock{tlsport = Port}) ->
+    tls_get_peer_finished_nif(Port);
+get_tls_last_message(self, #tlssock{tlsport = Port}) ->
+    tls_get_finished_nif(Port).
 
+-spec get_verify_result(tls_socket()) -> byte().
 get_verify_result(#tlssock{tlsport = Port}) ->
     {ok, Res} = get_verify_result_nif(Port),
     Res.
@@ -455,9 +424,6 @@ cert_verify_code(50) ->
 cert_verify_code(X) ->
     <<"Unknown OpenSSL error code: ", (integer_to_binary(X))/binary>>.
 
-integer_to_binary(I) ->
-    list_to_binary(integer_to_list(I)).
-
 encode_alpn(ProtoList) ->
     [<<(size(Proto)), Proto/binary>> || Proto <- ProtoList, Proto /= <<>>].
 
@@ -469,7 +435,9 @@ load_nif(SOPath) ->
     case erlang:load_nif(SOPath, 0) of
         ok ->
             ok;
-        {error, already_loaded} ->
+        {error, {reload, _}} -> % We don't support upgrade in this module so let's not crash
+            ok;
+        {error, {upgrade, _}} -> % We don't support upgrade in this module so let's not crash
             ok;
         {error, ErrorDesc} = Err ->
             error_logger:error_msg("failed to load TLS NIF: ~s~n",
@@ -479,37 +447,68 @@ load_nif(SOPath) ->
 
 -ifdef(TEST).
 
-load_nif_test() ->
-    SOPath = p1_nif_utils:get_so_path(fast_tls, [], "fast_tls"),
-    ?assertEqual(ok, load_nif(SOPath)).
+transmission_with_client_certificate_test() ->
+    transmission_test_with_opts([certificate()], [certificate()]).
 
-transmission_test() ->
-    {LPid, Port} = setup_listener([]),
-    SPid = setup_sender(Port, []),
+transmission_without_client_certificate_test() ->
+    transmission_test_with_opts([certificate()], []).
+
+transmission_without_server_cert_fails_test() ->
+    TestPid = self(),
+    {ok, ListenSocket} = gen_tcp:listen(0, [binary, {packet, 0}, {active, false},
+                                            {reuseaddr, true}, {nodelay, true}]),
+    {ok, Port} = inet:port(ListenSocket),
+    _ListenerPid = spawn(fun() -> {ok, Socket} = gen_tcp:accept(ListenSocket),
+                                 Res = tcp_to_tls(Socket, []),
+                                 TestPid ! {listener_tcp_to_tls, Res}
+                        end),
+    {ok, Socket} = gen_tcp:connect({127, 0, 0, 1}, Port, [binary, {packet, 0}, {active, false},
+                                                          {reuseaddr, true}, {nodelay, true}]),
+    {ok, TLSSock} = tcp_to_tls(Socket, [connect]),
+    close(TLSSock),
+    receive
+        {listener_tcp_to_tls, Res} ->
+            ?assertEqual({error, no_certfile}, Res)
+    end.
+
+
+transmission_test_with_opts(ListenerOpts, SenderOpts) ->
+    {LPid, Port} = setup_listener(ListenerOpts),
+    SPid = setup_sender(Port, SenderOpts),
     SPid ! {stop, self()},
-    receive
-        {result, Res} ->
-            ?assertEqual(ok, Res)
-    end,
+    FC = receive
+             {result, Res, FinishedFromClient} ->
+                 ?assertEqual(ok, Res),
+                 FinishedFromClient
+         end,
     LPid ! {stop, self()},
+    FL = receive
+             {received, Msg, FinishedFromListener} ->
+                 ?assertEqual(<<"abcdefghi">>, Msg),
+                 FinishedFromListener
+         end,
+    ?assertEqual(FC, FL),
     receive
-        {received, Msg} ->
-            ?assertEqual(<<"abcdefghi">>, Msg)
+        {certfile, Cert} ->
+            case lists:keymember(certfile, 1, SenderOpts) of
+                true -> ?assertNotEqual(error, Cert);
+                false -> ?assertEqual(error, Cert)
+            end
     end.
 
 not_compatible_protocol_options_test() ->
-    {LPid, Port} = setup_listener([{protocol_options, <<"no_sslv2|no_sslv3|no_tlsv1_1|no_tlsv1_2|no_tlsv1_3">>}]),
+    {LPid, Port} = setup_listener([certificate(), {protocol_options, <<"no_sslv2|no_sslv3|no_tlsv1_1|no_tlsv1_2|no_tlsv1_3">>}]),
     SPid = setup_sender(Port, [{protocol_options, <<"no_sslv2|no_sslv3|no_tlsv1|no_tlsv1_2|no_tlsv1_3">>}]),
     SPid ! {stop, self()},
     receive
-        {result, Res} ->
+        {result, Res, _} ->
             ?assertMatch({badmatch, {error, _}}, Res)
     end,
     LPid ! {stop, self()},
     receive
-        {received, {error, _, _} = Msg} ->
+        {received, {error, _, _} = Msg, _} ->
             ?assertMatch({error, _, <<>>}, Msg);
-        {received, Msg} ->
+        {received, Msg, _} ->
             ?assertMatch(<<>>, Msg)
     end.
 
@@ -519,30 +518,35 @@ setup_listener(Opts) ->
                                          {reuseaddr, true}, {nodelay, true}]),
     Pid = spawn(fun() ->
         {ok, Socket} = gen_tcp:accept(ListenSocket),
-        {ok, TLSSock} = tcp_to_tls(Socket, [{certfile, <<"../tests/cert.pem">>} | Opts]),
+        {ok, TLSSock} = tcp_to_tls(Socket, Opts),
         listener_loop(TLSSock, <<>>)
                 end),
     {ok, Port} = inet:port(ListenSocket),
     {Pid, Port}.
 
 listener_loop(TLSSock, Msg) ->
+    Finished = get_tls_last_message(peer, TLSSock),
     case recv(TLSSock, 1, 1000) of
         {error, timeout} ->
             receive
                 {stop, Pid} ->
-                    Pid ! {received, Msg}
+                    Pid ! {received, Msg, Finished},
+                    Cert = get_peer_certificate(TLSSock),
+                    Pid ! {certfile, Cert}
             after 0 ->
                 listener_loop(TLSSock, Msg)
             end;
         {error, closed} ->
             receive
                 {stop, Pid} ->
-                    Pid ! {received, Msg}
+                    Pid ! {received, Msg, Finished},
+                    Cert = get_peer_certificate(TLSSock),
+                    Pid ! {certfile, Cert}
             end;
         {error, Err} ->
             receive
                 {stop, Pid} ->
-                    Pid ! {received, {error, Err, Msg}}
+                    Pid ! {received, {error, Err, Msg}, Finished}
             end;
         {ok, Data} ->
             listener_loop(TLSSock, <<Msg/binary, Data/binary>>)
@@ -553,13 +557,14 @@ setup_sender(Port, Opts) ->
         binary, {packet, 0}, {active, false},
         {reuseaddr, true}, {nodelay, true}]),
     spawn(fun() ->
-        {ok, TLSSock} = tcp_to_tls(Socket, [connect, {certfile, <<"../tests/cert.pem">>} | Opts]),
+        {ok, TLSSock} = tcp_to_tls(Socket, [connect | Opts]),
         sender_loop(TLSSock)
           end).
 
 sender_loop(TLSSock) ->
-    Res = try
+    {Res, Finished} = try
               recv(TLSSock, 0, 100),
+              F = get_tls_last_message(self, TLSSock),
               ok = send(TLSSock, <<"abc">>),
               recv(TLSSock, 0, 100),
               ok = send(TLSSock, <<"def">>),
@@ -567,15 +572,18 @@ sender_loop(TLSSock) ->
               ok = send(TLSSock, <<"ghi">>),
               recv(TLSSock, 0, 100),
               close(TLSSock),
-              ok
+              {ok, F}
           catch
               _:Err ->
                   close(TLSSock),
-                  Err
+                  {Err, <<>>}
           end,
     receive
         {stop, Pid} ->
-            Pid ! {result, Res}
+            Pid ! {result, Res, Finished}
     end.
 
+certificate() ->
+    {certfile, <<"../tests/cert.pem">>}.
+
 -endif.
